@@ -20,7 +20,7 @@ INJECT_FUNCTION_TOKEN = "INJECT_FUNCTION"
 
 # Suffixes to add to the base prompt.
 STACK_TRACE_SUFFIX = "\n\nSTACK TRACE: "
-RETRY_SUFFIX = "\n\nAttempting again.\n\nOUTPUT: "
+RETRY_SUFFIX = "\n\nAttempting again.\n\nOUTPUT: ```python"
 
 # Prompts! The best part :).
 BASE_PROMPT = """You have an instance `env` with methods:
@@ -34,18 +34,18 @@ BASE_PROMPT = """You have an instance `env` with methods:
 - `env.wait(seconds)` waits for `seconds`.
 - `env.scroll(direction, iframe=None)` scrolls. Switches to `iframe` if given. `direction` can be "up", "down", "left", or "right". 
 - `env.get_llm_response(text)` asks AI about a string `text`.
+- `env.query_memory(prompt)` asks AI to query its memory of ALL the web pages it has browsed so far. Invoked with something like "Query memory".
 - `env.retrieve_information(prompt, entire_page=False)` returns a string, information from a page given a prompt. Use prompt="Summarize:" for summaries. Uses all the text if entire_page=True and only visible text if False. Invoked with commands like "retrieve", "find in the page", or similar.
 - `env.ask_llm_to_find_element(description)` asks AI to find an WebElement that matches the description. It returns None if it cannot find an element that matches the description, so you must check for that.
-- `env.query_memory(prompt)` If enable_memory=True, then asks AI to query its memory (an embeddings index) of the web pages it has browsed. Invoked with something like "Query memory".
 - `env.screenshot(element, filename)` takes a screenshot of the element and saves it to `filename`.
 - `env.save(text, filename)` saves the string `text` to a file `filename`.
 - `env.get_text_from_page(entire_page)` returns the free text from the page. If entire_page is True, it returns all the text from HTML doc. If False, returns only visible text.
 
-WebElement has functions:
+Guidelines for using GPTWebElement:
 1. `element.text` returns the text of the element.
 2. `element.get_attribute(attr)` returns the value of the attribute of the element. If the attribute does not exist, it returns ''.
 3. `element.find_elements(by='id', value=None)` is similar to `env.find_elements()` except that it only searches the children of the element and does not search iframes.
-4. `element.is_displayed()` returns if the element is visible.
+4. `env.is_element_visible_in_viewport(element)` returns if the element is visible in the viewport.
 5. Do NOT use `element.send_keys(text)` or `element.click()`. Use `env.send_keys(text)` and `env.click(element)` instead.
 
 In xpaths, to get the text of an element, do NOT use `text()`. Use `normalize-space()` instead.
@@ -115,11 +115,17 @@ class InstructionCompiler:
         self.prompt_to_find_element = PROMPT_TO_FIND_ELEMENT
         self.use_compiled = use_compiled
         self.api_cache = {}  # Instruction string to API response.
+        self.functions = {}  # Set in _parse_instructions_into_queue.
+        self.finished_instructions = []
+        self.history = []  # Keep track of the history of actions.
 
-        # Load instructions.
-        # - Initialize self.instructions to be a dict with the key
-        #   "instructions", and possibly "compiled" and "chrome_options".
-        # - Initialize instructions_str for the queue.
+        # Set the instructions.
+        self.instructions = instructions  # Overriden in set_instructions.
+        self.compiled_instructions = []  # Overriden if available.
+        self.instructions_queue = []  # Overriden in set_instructions.
+        self.set_instructions(instructions)
+
+    def set_instructions(self, instructions: Union[str, dict, io.TextIOWrapper]):
         self.instructions = self._load_instructions(instructions)
         self.compiled_instructions = []
         if isinstance(self.instructions, str):
@@ -136,10 +142,7 @@ class InstructionCompiler:
         else:
             raise ValueError("Instructions must be either a string or a dict.")
 
-        self.functions = {}  # Set in _parse_instructions_into_queue.
         self.instructions_queue = self._parse_instructions_into_queue(instructions_str)
-        self.finished_instructions = []
-        self.history = []  # Keep track of the history of actions.
 
     def _load_instructions(
         self, instructions: Union[str, dict, io.TextIOWrapper]
@@ -253,7 +256,9 @@ class InstructionCompiler:
 
         return final_queue
 
-    def get_completion(self, prompt, temperature=0, model=None, use_cache=True):
+    def get_completion(
+        self, prompt, model=None, temperature=0, max_tokens=1024, use_cache=True
+    ):
         """Wrapper over OpenAI's completion API."""
         if model is None:
             model = self.model
@@ -269,7 +274,7 @@ class InstructionCompiler:
                 response = openai.ChatCompletion.create(
                     model=model,
                     messages=[{"role": "user", "content": prompt}],
-                    max_tokens=512,
+                    max_tokens=max_tokens,
                     top_p=1,
                     frequency_penalty=0,
                     presence_penalty=0,
@@ -281,7 +286,7 @@ class InstructionCompiler:
                 response = openai.Completion.create(
                     model=model,
                     prompt=prompt,
-                    max_tokens=512,
+                    max_tokens=max_tokens,
                     top_p=1,
                     frequency_penalty=0,
                     presence_penalty=0,
@@ -290,14 +295,20 @@ class InstructionCompiler:
                     stop=["```"],
                 )
                 text = response["choices"][0]["text"]
-        except openai.error.RateLimitError as exc:
+        except (
+            openai.error.RateLimitError,
+            openai.error.APIError,
+            openai.error.Timeout,
+        ) as exc:
             logger.info(
-                "Rate limit error: {exc}. Sleeping for a few seconds.".format(
+                "OpenAI error. Likely a rate limit error, API error, or timeout: {exc}. Sleeping for a few seconds.".format(
                     exc=str(exc)
                 )
             )
             time.sleep(5)
-            text = self.get_completion(prompt, temperature, model)
+            text = self.get_completion(
+                prompt, temperature=temperature, max_tokens=max_tokens, model=model
+            )
 
         # Add to cache.
         self.api_cache[prompt] = text
