@@ -6,31 +6,29 @@ import sys
 import time
 import traceback
 import html2text
-import nltk
-from nltk.tokenize import sent_tokenize
 from bs4 import BeautifulSoup
 from bs4.element import NavigableString
 from bs4.element import Tag
-from llama_index import Document, GPTSimpleVectorIndex
+from llama_index import Document, GPTSimpleVectorIndex, LLMPredictor
 from selenium import webdriver
 from selenium.webdriver.chrome.service import Service
 from selenium.webdriver.common.by import By
 from selenium.webdriver.common.keys import Keys
 from selenium.webdriver.common.action_chains import ActionChains
 from selenium.webdriver.support.relative_locator import locate_with
+from langchain.chat_models import ChatOpenAI
 from .compilers.instruction_compiler import InstructionCompiler
 from .memories import Memory
 
 TIME_BETWEEN_ACTIONS = 0.1
 
-nltk.download("punkt")
-
 import logging
 
 logging.basicConfig(level=logging.INFO)
 logger = logging.getLogger(__name__)
 
 NO_RESPONSE_TOKEN = "<NONE>"  # To denote that empty response from model.
+CHATGPT_KWARGS = kwargs = {"temperature": 0, "model_name": "gpt-3.5-turbo"}
 
 
 class GPTWebElement(webdriver.remote.webelement.WebElement):
@@ -59,6 +57,7 @@ class GPTSeleniumAgent:
         debug=False,
         debug_html_folder="",
         instruction_output_file=None,
+        close_after_completion=True,
     ):
         """Initialize the agent.
 
@@ -71,7 +70,7 @@ class GPTSeleniumAgent:
             model_for_instructions (str): OpenAI model to use for generating
                 instructions.
             model_for_responses (str): OpenAI model to use for generating
-                responses from `ask_llm_for_response`.
+                responses from `get_llm_response`.
             user_data_dir (str): Path to the user data directory created by
                 Selenium.
             headless (bool): Whether to run the browser in headless mode.
@@ -82,6 +81,8 @@ class GPTSeleniumAgent:
                 should be saved.
             instruction_output_file (str): Path to the YAML file where the
                 instructions should be saved.
+            close_after_completion (bool): Whether to close the browser after
+                the instructions have been executed.
         """
         """Helpful instance variables."""
         assert (
@@ -96,6 +97,7 @@ class GPTSeleniumAgent:
         self.debug = debug
         self.debug_html_folder = debug_html_folder
         self.enable_memory = enable_memory
+        self.close_after_completion = close_after_completion
 
         """Fire up the compiler."""
         self.instruction_compiler = InstructionCompiler(
@@ -151,7 +153,7 @@ class GPTSeleniumAgent:
 
         return False
 
-    def __remove_blacklisted_elements_and_attributes(self):
+    def _remove_blacklisted_elements_and_attributes(self) -> BeautifulSoup:
         """Clean HTML to remove blacklisted elements and attributes. Returns
         BeautifulSoup object."""
         blacklisted_elements = set(
@@ -200,7 +202,7 @@ class GPTSeleniumAgent:
         First removes blacklisted elements and attributes, then removes any
         children of elements. Finally, removes any elements with no attrs.
         """
-        soup = self.__remove_blacklisted_elements_and_attributes()
+        soup = self._remove_blacklisted_elements_and_attributes()
         # Remove children of elements that have children.
         elements = soup.find_all()
         [ele.clear() if ele.contents else ele for ele in elements if ele.contents]
@@ -216,7 +218,9 @@ class GPTSeleniumAgent:
             exec(instructions, globals(), ldict)
         except:
             self.__handle_agent_exception(instructions)
-        exec("env.driver.quit()", globals(), ldict)
+
+        if self.close_after_completion:
+            self.driver.quit()
 
     def __print_instruction_and_action(self, instruction, action):
         """Logging the instruction and action."""
@@ -287,6 +291,7 @@ class GPTSeleniumAgent:
                 self.__save_html_snapshot()
 
             logger.info(traceback.print_exc())
+            logger.info("Starting interactive debugger. Type `env` for the Agent object.")
             env = self  # For the interactive debugger.
             pdb.set_trace()
 
@@ -330,7 +335,8 @@ class GPTSeleniumAgent:
                 self.instruction_output_file
             )
 
-        exec("env.driver.quit()", globals(), ldict)
+        if self.close_after_completion:
+            self.driver.quit()
 
     def __switch_to_element_iframe(func):
         """Decorator function to switch to the iframe of the element."""
@@ -338,17 +344,25 @@ class GPTSeleniumAgent:
         def wrapper(*args):
             self = args[0]
             element = args[1]
-            if element is not None:
+            if isinstance(element, GPTWebElement) and (element is not None):
                 iframe = element.iframe
                 if iframe is not None:
                     self.driver.switch_to.frame(iframe)
-                func(*args)
+                result = func(*args)
                 self.driver.switch_to.default_content()
+            else:
+                result = func(*args)
+            
+            return result
 
         return wrapper
 
     """Functions meant for the client to call."""
 
+    def set_instructions(self, instructions):
+        """Reset the instructions to `instructions`."""
+        self.instruction_compiler.set_instructions(instructions)
+
     def run(self):
         """Run the agent."""
         should_use_compiled = self.instruction_compiler.use_compiled
@@ -377,6 +391,23 @@ class GPTSeleniumAgent:
             text = self.get_text_from_page(entire_page=False)
             self.memory.add(text)
 
+    @__switch_to_element_iframe
+    def is_element_visible_in_viewport(self, element: GPTWebElement) -> bool:
+        is_visible = self.driver.execute_script(
+            "var elem = arguments[0],                 " 
+            "  box = elem.getBoundingClientRect(),    " 
+            "  cx = box.left + box.width / 2,         " 
+            "  cy = box.top + box.height / 2,         " 
+            "  e = document.elementFromPoint(cx, cy); " 
+            "for (; e; e = e.parentElement) {         " 
+            "  if (e === elem)                        " 
+            "    return true;                         " 
+            "}                                        " 
+            "return false;                            ",
+            element
+        )
+        return is_visible
+
     def scroll(self, direction=None, iframe=None):
         assert direction in ["up", "down", "left", "right"]
         assert (iframe is None) or isinstance(iframe, GPTWebElement)
@@ -519,32 +550,26 @@ class GPTSeleniumAgent:
     def retrieve_information(self, prompt, entire_page=False):
         """Retrieves information using using GPT-Index embeddings from a page."""
         text = self.get_text_from_page(entire_page=entire_page)
-
-        # Tokenize by sentence, and then load each set of three sentences as
-        # a doc.
-        sentences = sent_tokenize(text)
-        docs = []
-        for i in range(0, len(sentences), 5):
-            doc = " ".join(sentences[i : i + 5])
-            docs.append(Document(doc))
-
-        # Then we use GPT Index to summarize the text.
+        index = GPTSimpleVectorIndex([Document(text)])
         logger.info(
-            "Found {num_docs} documents for indexing.".format(num_docs=len(docs))
+            'Retrieving information from web page with prompt: "{prompt}"'.format(prompt=prompt)
         )
-        index = GPTSimpleVectorIndex(docs)
-        print(text[:150])
-        logger.info(
-            'Retrieving information with prompt: "{prompt}"'.format(prompt=prompt)
+        resp = index.query(
+            prompt,
+            llm_predictor=LLMPredictor(llm=ChatOpenAI(**CHATGPT_KWARGS))
         )
-        resp = index.query(prompt, similarity_top_k=3)
         return resp.response.strip()
 
     def get_llm_response(self, prompt, temperature=0.7, model=None):
         if model is None:
             model = self.model_for_responses
 
-        return self.instruction_compiler.get_completion(prompt, temperature, model)
+        return self.instruction_compiler.get_completion(
+            prompt,
+            model=model,
+            max_tokens=2048,  # Let it be expressive!
+            temperature=temperature,
+        )
 
     def query_memory(self, prompt):
         """Queries the memory of the LLM."""
@@ -586,7 +611,8 @@ class GPTSeleniumAgent:
         query = "Find element that matches description: {element_description}. If no element matches, return {no_resp_token}.".format(
             element_description=element_description, no_resp_token=NO_RESPONSE_TOKEN
         )
-        resp = index.query(query)
+        query = query + " Please be as succinct as possible, with no additional commentary."
+        resp = index.query(query, llm_predictor=LLMPredictor(llm=ChatOpenAI(**CHATGPT_KWARGS)))
         doc_id = resp.source_nodes[0].doc_id
 
         resp_text = resp.response.strip()
